from ..errors import orm as orm_errors

import threading


class ConnPoolRef:
    def __init__(self, thread: threading.Thread, conn, cursor, cleanup: callable):
        self.thread = thread
        self.conn = conn
        self.cursor = cursor
        self.cleanup = cleanup

    def __del__(self):
        self.cleanup()


class ConnPool:
    def __init__(self, driver, db_name: str, **conn_ops):
        self.driver = driver
        self.conn_info = (db_name, conn_ops)
        self.pool = {}

    def __del__(self):
        if len(self.pool) > 0:
            self.close()
            raise orm_errors.connection_pool.ConnPoolNotClosedError(
                "The connection pool is not fully closed. Automatically closed all connections for you."
            )

    def _new_conn(self):
        """
        [Private] Create a new connection

        .. warning::
            THIS IS A PRIVATE METHOD, DO NOT CALL IT DIRECTLY OR IT WILL CAUSE UNEXPECTED BEHAVIOR
        """
        db_aname, conn_ops = self.conn_info
        conn = self.driver.connect(db_aname, **conn_ops)
        return conn

    def _create_conn(self, thread: threading.Thread):
        """
        [Private] Create a new connection for the target thread, and add it to the pool

        .. warning::
            THIS IS A PRIVATE METHOD, DO NOT CALL IT DIRECTLY OR IT WILL CAUSE UNEXPECTED BEHAVIOR
        """
        # ensure the id is unique and predictable
        conn_id = id(thread)

        conn = self._new_conn()
        cursor = conn.cursor()

        # TODO: We can't close the connection in this way.
        #       Because the conn is created in the 'Target Thread', while the _remove_conn
        #       method will be called in the 'Monitor Thread' when 'Target Thread' is over.
        #       So the conn can only be closed in the 'Monitor Thread', which is not allow-
        #       ed in sqlite3.
        #       I've tried the `threading.local()`, but its also executed in the Main thread.
        self.pool[conn_id] = ConnPoolRef(
            thread, conn, cursor, lambda: self._remove_conn(conn_id)
        )
        t_local = threading.local()
        t_local.conn_ref = self.pool[conn_id]

    def _remove_conn(self, conn_id):
        """
        [Private] Remove the connection from the pool

        .. warning::
            THIS IS A PRIVATE METHOD, DO NOT CALL IT DIRECTLY OR IT WILL CAUSE UNEXPECTED BEHAVIOR
        """
        ref = self.pool.pop(conn_id)
        ref.conn.close()

    def get_ref(self) -> ConnPoolRef:
        """
        Get the ConnPoolRef object for the current thread.
        Create a new connection if the current thread is not in the pool.

        :return: The ConnPoolRef object
        :rtype: ConnPoolRef
        """
        thread = threading.current_thread()
        conn_id = id(thread)

        if conn_id not in self.pool:
            self._create_conn(thread)

        return self.pool[conn_id]

    def get_conn(self):
        """
        Get the connection object for the current thread
        Create a new connection if the current thread is not in the pool.
        """
        return self.get_ref().conn

    def get_cursor(self):
        """
        Get the cursor object for the current thread
        Create a new connection if the current thread is not in the pool.
        """
        return self.get_ref().cursor
    
    def commit(self):
        """
        Commit all the current thread's connections
        """
        conn = self.get_conn()
        conn.commit()

    def close_all(self):
        """
        Close all the connections in the pool
        """
        for conn_id in list(self.pool.keys()):
            self._remove_conn(conn_id)



class ConnPoolErrors(Exception):
    """
    The super class of all errors related to the connection pool
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

class ConnPoolNotClosedError(ConnPoolErrors):
    """
    A connection pool is not fully closed
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


"""
Copyright (c) Bernie Huang 2023, all rights reserved.

MercurySQL.gensql.database
==========================
This file contains the DataBase class and a tool function `set_driver` as well.

Classes
-------
- `DataBase`: Represents a SQL database and provides methods for creating tables, executing SQL commands, and retrieving table objects.

Methods
-------
- `set_driver`: Set the default driver for the `DataBase` class.
"""
from typing import List

from ..orm.command_queue import CommandQueue
from ..drivers import BaseDriver
from ..errors import *

from .table import Table


# ========== Class Decorations ==========
class DataBase:
    pass


# ========= Collect Infos =========
import os

real_path = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(real_path, "..", "VERSION"), encoding="utf-8") as f:
    __version__ = f.read().strip()


# ========= Tool Functions =========
default_driver = None


def set_driver(driver):
    """
    Set the default driver for the database, so you won't need to specify the driver for each DB every time.
    """
    global default_driver
    default_driver = driver


def check_version(version: str) -> bool:
    """
    Check whether the version is supported.

    :param version: The version to check.
    :type version: str

    :return: Whether the version is supported.
    :rtype: bool
    """
    cv = list(map(int, __version__.split(".")))  # current version
    dv = list(map(int, version.split(".")))  # driver version

    vdiff = [cv[i] - dv[i] for i in range(len(cv))]

    # major version
    if vdiff[0] != 0:
        return False
    # minor version (driver is older)
    if vdiff[1] > 0:
        return False

    return True


# ========== Classes ==========
class DataBase:
    """
    Select/Create/Connect a SQL database.
    Represents a SQL database and provides methods for creating tables, executing SQL commands, and retrieving table objects.
    """

    def __init__(self, db_name: str, driver=None, **kwargs):
        """
        Create a new database object.

        :param db_name: The name of the database.
        :type db_name: str

        Example Usage:

        .. code-block:: python

            db = DataBase('test.db')

        How It Works:
            - start a connection to the SQL database, using the driver specified by `driver` parameter or `set_driver()` method.
            - get a cursor to execute sql commands.
            - gather all infomations of the database, for further usages.
        """
        if driver is None:
            driver = default_driver

        if not issubclass(driver, BaseDriver):
            raise NotSpecifiedError("Driver not specified.")
        else:
            self.driver = driver

        if not check_version(driver.version):  # check version
            raise DriverIncompatibleError(driver.__name__, driver.version, __version__)

        # self.driver = driver()  # normal way
        # self.conn_pool = ConnPool(self.driver, db_name, **kwargs)  # connection pool
        self.cq = CommandQueue(driver, db_name, **kwargs)
        self.conn = driver.connect(db_name, **kwargs)
        # self.cursor = self.conn.cursor()  # normal way

        self.info = {"name": db_name}

        self.template = None
        self.template_params = {}

        self._gather_info()

    def _gather_info(self):
        """
        [Helper]
        Gather all infomations of the database, including:
          - all tables
        """
        self.tables = self.driver.APIs.get_all_tables(self.conn)
        self.tables = {tname: Table(self, tname) for tname in self.tables}

    def do(self, *sql: str, paras: List[tuple] = []):
        """
        Execute a sql command on the database.

        :param sql: The sql command(s).
        :type sql: str
        :param paras: The parameters for the sql command(s).
        :type paras: List[tuple]

        :return: The cursor of the database.
        :rtype: Driver.Cursor

        Example Usage:

        .. code-block:: python

            db = DataBase('test.db')
            db.do("CREATE TABLE test (id INTEGER PRIMARY KEY, name TEXT)")
            db.do("INSERT INTO test (name) VALUES (?)", paras=[('Bernie',)])
            db.do(
                "SQL1",
                "SQL2",
                paras=[
                       (paras1, paras2),
                       (paras3, paras4)
                ])

        How It Works:
            - execute sql commands one by one, with parameters
            - commit after all commands are executed
        """
        # Recommended to use `db.do(sql1, sql2)` instead of `db.do([sql1, sql2])`.
        if isinstance(sql[0], list):
            sql = sql[0]

        if len(paras) < len(sql):
            paras += [()] * (len(sql) - len(paras))

        # start a new cursor
        # c = self.conn.cursor()    # normal way
        # c = self.conn_pool.get_cursor()   # connection pool
        c = self.cq.get_cursor()

        # for each sql command
        for i in range(len(sql)):
            cmd = sql[i].replace(
                "___!!!PAYLOAD!!!___", self.driver.payload
            )  # replace payload

            c.execute(cmd, paras[i])

        # commit changes
        # try:
            # self.conn.commit()    # normal way
            # self.conn_pool.commit()   # connection pool
        # except:
            # pass

        return c

    def setTemplate(self, template: dict, **kwargs) -> None:
        """
        Set the template, so new table's structure will be setted to this template.
        You can specifd the template by using `template=...` parameter into `db.createTable()` method.
        Or set it after the table is created, using `table.struct()`.

        :param template: The template to set.
        :type template: dict
        :param \*\*kwargs: The parameters for the template. E.g., `primaryKey='id'`.

        Example Usage:

        .. code-block:: python

            db = DataBase('test.db')
            db.setTemplate({
                'id': int,
                'name': str
            })
            table = db.createTable('test')

        """
        self.template = template
        self.template_params = kwargs

    def createTable(
        self, *table_names: str, force: bool = False, template=None
    ) -> Table:
        """
        create a table in the database.

        :param table_names: The name of the table.
        :type table_names: str
        :param force: Allow to return an existing table.
        :type force: bool

        :return: A Table object.
        :rtype: Table

        Example Usage:

        .. code-block:: python

            db = DataBase('test.db')
            table = db.createTable('test')

        How It Works:
            - create table(s) if not exists
            - if already exists, return the existing table(s) in a NEW `Table` Object.
            - if exists and `force` set to False, raise an Exception.
        """
        # using template in `self.template` if not specified
        if template is None:
            template = self.template

        tables = []
        already_exists = False

        for table_name in table_names:
            if table_name in self.tables:
                already_exists = True
                if not force:
                    raise DuplicateError(f"Table `{table_name}` already exists.")

            table = Table(self, table_name)

            # set template
            if not already_exists and template is not None:
                table.struct(template, **self.template_params)

            # must be executed after `_gather_into()`, because the following code will use a mapping between table name and table object
            self.tables[table_name] = table

            tables.append(table)

        return tables if len(tables) > 1 else tables[0]

    def __getitem__(self, key: str) -> Table:
        """
        Create / Choose a table from the database.

        :param key: The name of the table.
        :type key: str

        :return: A Table object.
        :rtype: Table

        Example Usage:

        .. code-block:: python

            db = DataBase('test.db')
            table = db['test']

        How It Works:
            - if exists, return the existing table (the OLD `Table` Object)
            - if not exists, create a new table by `createTable()`.

        .. note:: The only difference between `__getitem__()` and `createTable()` is that `__getitem__()` will return the **OLD** `Table` Object if exists, while `createTable()` will return a **NEW** `Table` Object.
        """
        if key in self.tables:
            return self.tables[key]
        else:
            return self.createTable(key, force=True)

    def deleteTable(self, *table_names: str) -> None:
        """
        delete a table in the database.

        :param table_names: The name of the table.
        :type table_names: str

        Example Usage:

        .. code-block:: python

            db = DataBase('test.db')
            db.deleteTable('test')

        How It Works:
            - delete table(s) if exists
            - raise an Exception if not exists.
        """
        for table_name in table_names:
            if table_name not in self.tables:
                raise NotExistsError(f"Table `{table_name}` not exists.")

            cmd = self.driver.APIs.gensql.drop_table(table_name)
            self.do(cmd)

            del self.tables[table_name]

    def __delitem__(self, key: str) -> None:
        """
        Delete a table from the database, same as `deleteTable()`.

        :param key: The name of the table.
        :type key: str

        Example Usage:

        .. code-block:: python

            db = DataBase('test.db')
            del db['test']  # same as db.deleteTable('test')

        """
        self.deleteTable(key)

    # def __del__(self):    # connection pool
    #     """
    #     Close the connection when the object is deleted.
    #     """
    #     try:
    #         self.conn_pool.close_all()
    #         # self.conn.close()
    #     except AttributeError:
    #         # Not initialized
    #         pass



"""
This file offers the `CommandQueue` class, which is used to queue and execute commands in the MercurySQL package.

This is useful when you are working with multiple threads and want to execute commands in a specific order.
"""

import queue
import threading

FETCHALL = -1


class CommandQueue:
    def __init__(self, driver, db_name: str, **conn_ops):
        self.driver = driver
        self.conn_info = (db_name, conn_ops)

        self.queue = queue.Queue()

        # start the loop, starts to process the project now
        self.isRunning = True
        self.loop()

    def stop(self):
        self.isRunning = False

    def __del__(self):
        self.stop()

    def put(self, command):
        """
        Put a command into the queue.

        :param command: The command to put into the queue.
        :type command: Any
        """
        self.queue.put(command)

    def loop(self):
        """
        Loop through the queue and execute each command (in an new thread).
        When the queue is empty, the loop will wait for new commands.
        """

        # start a new thread
        def loop_thread():
            # start a new conn
            db_name, conn_ops = self.conn_info
            conn = self.driver.connect(db_name, **conn_ops)
            cursor = conn.cursor()

            while self.isRunning:
                query, param, callback = self.queue.get()

                cursor.execute(query, param)
                result = cursor.fetchall()

                callback(result)
                conn.commit()
                self.queue.task_done()
            
            cursor.close()
            conn.close()

        # Start a Daemon thread, which will be killed when the main thread is over.
        threading.Thread(target=loop_thread, daemon=True).start()

    def get_cursor(self):
        """
        Return a fake cursor that will transfer the command to the real cursor in seperate thread.

        :return: The cursor.
        :rtype: Cursor
        """

        return CQFakeCursor(self)


class CQFakeCursor:
    def __init__(self, cq):
        self.cq = cq
        self.event = threading.Event()

    def execute(self, query: str, param: tuple) -> None:
        """
        Execute a query.

        :param query: The query to execute.
        :type query: str
        :param param: The parameters for the query.
        :type param: tuple
        """
        def callback(result):
            self.result = result
            self.event.set()

        self.cq.put((query, param, callback))
        self.event.wait()  # wait

        return None

    def fetchall(self):
        """
        Fetch all results.

        :return: The results.
        :rtype: list
        """
        return self.result

    def fetchone(self):
        """
        Fetch one result.

        :return: The result.
        :rtype: Any
        """
        return self.result[0]

    def fetchmany(self, size: int):
        """
        Fetch a specific number of results.

        :param size: The number of results to fetch.
        :type size: int
        """
        return self.result[:size]



class DuplicateError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)


class NotExistsError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)


class NotSupportedError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)


class ConfilictError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)

class NotSpecifiedError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)



class DriverIncompatibleError(Exception):
    def __init__(self, driver: str, dv: str, cv: str):
        """
        :param driver: Driver Name.
        :param dv: Driver Version.
        :param cv: Current Version.
        """
        self.message = f"Driver `{driver}`(v{dv}) is not supported by MercurySQL(v{cv}). Please update your Driver/MercurySQL to the latest version."
        super().__init__(self.message)



from setuptools import setup, find_packages

# Read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


with open('MercurySQL/VERSION', 'r') as f:
    version = f.read().strip()

setup(
    name='MercurySQL',
    version=version,
    description='Operate sql in a more pythonic way.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Bernie J. Huang',
    author_email='berniehuang2008@163.com',
    url='https://github.com/BernieHuang2008/MercurySQL',
    packages=find_packages(),
    package_data={
        'MercurySQL': ['VERSION'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
    ],
)
